@@ -11,7 +11,6 @@
 //     go run linux/mkall.go <linux_dir> <glibc_dir>
 
 //go:build ignore
-// +build ignore
 
 package main
 
@@ -24,7 +23,6 @@ import (
 	"fmt"
 	"go/build/constraint"
 	"io"
-	"io/ioutil"
 	"os"
 	"os/exec"
 	"path/filepath"
@@ -630,11 +628,7 @@ func (t *target) matchesMksyscallFile(file string) (bool, error) {
 	s := bufio.NewScanner(f)
 	for s.Scan() {
 		// Keep scanning until a valid constraint is found or we hit EOF.
-		//
-		// This only supports single-line constraints such as the //go:build
-		// convention used in Go 1.17+. Because the old //+build convention
-		// (which may have multiple lines of build tags) is being deprecated,
-		// we don't bother looking for multi-line constraints.
+		// This is sufficient for the single-line //go:build constraints.
 		if expr, err = constraint.Parse(s.Text()); err == nil {
 			found = true
 			break
@@ -753,8 +747,6 @@ func generatePtracePair(arch1, arch2, archName string) error {
 	fmt.Fprintf(buf, "// Code generated by linux/mkall.go generatePtracePair(%q, %q). DO NOT EDIT.\n", arch1, arch2)
 	fmt.Fprintf(buf, "\n")
 	fmt.Fprintf(buf, "//go:build linux && (%s || %s)\n", arch1, arch2)
-	fmt.Fprintf(buf, "// +build linux\n")
-	fmt.Fprintf(buf, "// +build %s %s\n", arch1, arch2)
 	fmt.Fprintf(buf, "\n")
 	fmt.Fprintf(buf, "package unix\n")
 	fmt.Fprintf(buf, "\n")
@@ -810,7 +802,7 @@ func generatePtraceRegSet(arch string) error {
 // ptraceDef returns the definition of PtraceRegs for arch.
 func ptraceDef(arch string) (string, error) {
 	filename := fmt.Sprintf("ztypes_linux_%s.go", arch)
-	data, err := ioutil.ReadFile(filename)
+	data, err := os.ReadFile(filename)
 	if err != nil {
 		return "", fmt.Errorf("reading %s: %v", filename, err)
 	}
