@@ -36,8 +36,10 @@ import (
 	"google.golang.org/grpc/connectivity"
 	"google.golang.org/grpc/credentials"
 	"google.golang.org/grpc/credentials/insecure"
+	"google.golang.org/grpc/internal"
 	"google.golang.org/grpc/internal/balancer/stub"
 	"google.golang.org/grpc/internal/balancerload"
+	"google.golang.org/grpc/internal/grpcsync"
 	"google.golang.org/grpc/internal/grpcutil"
 	imetadata "google.golang.org/grpc/internal/metadata"
 	"google.golang.org/grpc/internal/stubserver"
@@ -85,6 +87,7 @@ func (b *testBalancer) UpdateClientConnState(state balancer.ClientConnState) err
 	// Only create a subconn at the first time.
 	if b.sc == nil {
 		var err error
+		b.newSubConnOptions.StateListener = b.updateSubConnState
 		b.sc, err = b.cc.NewSubConn(state.ResolverState.Addresses, b.newSubConnOptions)
 		if err != nil {
 			logger.Errorf("testBalancer: failed to NewSubConn: %v", err)
@@ -97,21 +100,17 @@ func (b *testBalancer) UpdateClientConnState(state balancer.ClientConnState) err
 }
 
 func (b *testBalancer) UpdateSubConnState(sc balancer.SubConn, s balancer.SubConnState) {
-	logger.Infof("testBalancer: UpdateSubConnState: %p, %v", sc, s)
-	if b.sc != sc {
-		logger.Infof("testBalancer: ignored state change because sc is not recognized")
-		return
-	}
-	if s.ConnectivityState == connectivity.Shutdown {
-		b.sc = nil
-		return
-	}
+	panic(fmt.Sprintf("UpdateSubConnState(%v, %+v) called unexpectedly", sc, s))
+}
+
+func (b *testBalancer) updateSubConnState(s balancer.SubConnState) {
+	logger.Infof("testBalancer: updateSubConnState: %v", s)
 
 	switch s.ConnectivityState {
 	case connectivity.Ready:
-		b.cc.UpdateState(balancer.State{ConnectivityState: s.ConnectivityState, Picker: &picker{sc: sc, bal: b}})
+		b.cc.UpdateState(balancer.State{ConnectivityState: s.ConnectivityState, Picker: &picker{bal: b}})
 	case connectivity.Idle:
-		b.cc.UpdateState(balancer.State{ConnectivityState: s.ConnectivityState, Picker: &picker{sc: sc, bal: b, idle: true}})
+		b.cc.UpdateState(balancer.State{ConnectivityState: s.ConnectivityState, Picker: &picker{bal: b, idle: true}})
 	case connectivity.Connecting:
 		b.cc.UpdateState(balancer.State{ConnectivityState: s.ConnectivityState, Picker: &picker{err: balancer.ErrNoSubConnAvailable, bal: b}})
 	case connectivity.TransientFailure:
@@ -125,7 +124,6 @@ func (b *testBalancer) ExitIdle() {}
 
 type picker struct {
 	err  error
-	sc   balancer.SubConn
 	bal  *testBalancer
 	idle bool
 }
@@ -135,14 +133,14 @@ func (p *picker) Pick(info balancer.PickInfo) (balancer.PickResult, error) {
 		return balancer.PickResult{}, p.err
 	}
 	if p.idle {
-		p.sc.Connect()
+		p.bal.sc.Connect()
 		return balancer.PickResult{}, balancer.ErrNoSubConnAvailable
 	}
 	extraMD, _ := grpcutil.ExtraMetadata(info.Ctx)
 	info.Ctx = nil // Do not validate context.
 	p.bal.pickInfos = append(p.bal.pickInfos, info)
 	p.bal.pickExtraMDs = append(p.bal.pickExtraMDs, extraMD)
-	return balancer.PickResult{SubConn: p.sc, Done: func(d balancer.DoneInfo) { p.bal.doneInfo = append(p.bal.doneInfo, d) }}, nil
+	return balancer.PickResult{SubConn: p.bal.sc, Done: func(d balancer.DoneInfo) { p.bal.doneInfo = append(p.bal.doneInfo, d) }}, nil
 }
 
 func (s) TestCredsBundleFromBalancer(t *testing.T) {
@@ -168,7 +166,9 @@ func (s) TestCredsBundleFromBalancer(t *testing.T) {
 
 	cc := te.clientConn()
 	tc := testgrpc.NewTestServiceClient(cc)
-	if _, err := tc.EmptyCall(context.Background(), &testpb.Empty{}); err != nil {
+	ctx, cancel := context.WithTimeout(context.Background(), defaultTestTimeout)
+	defer cancel()
+	if _, err := tc.EmptyCall(ctx, &testpb.Empty{}); err != nil {
 		t.Fatalf("Test failed. Reason: %v", err)
 	}
 }
@@ -195,14 +195,10 @@ func testPickExtraMetadata(t *testing.T, e env) {
 	te.startServer(&testServer{security: e.security})
 	defer te.tearDown()
 
-	// Set resolver to xds to trigger the extra metadata code path.
-	r := manual.NewBuilderWithScheme("xds")
-	resolver.Register(r)
-	defer func() {
-		resolver.UnregisterForTesting("xds")
-	}()
-	r.InitialState(resolver.State{Addresses: []resolver.Address{{Addr: te.srvAddr}}})
-	te.resolverScheme = "xds"
+	// Trigger the extra-metadata-adding code path.
+	defer func(old string) { internal.GRPCResolverSchemeExtraMetadata = old }(internal.GRPCResolverSchemeExtraMetadata)
+	internal.GRPCResolverSchemeExtraMetadata = "passthrough"
+
 	cc := te.clientConn()
 	tc := testgrpc.NewTestServiceClient(cc)
 
@@ -246,11 +242,11 @@ func testDoneInfo(t *testing.T, e env) {
 	cc := te.clientConn()
 	tc := testgrpc.NewTestServiceClient(cc)
 
-	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
+	ctx, cancel := context.WithTimeout(context.Background(), defaultTestTimeout)
 	defer cancel()
 	wantErr := detailedError
 	if _, err := tc.EmptyCall(ctx, &testpb.Empty{}); !testutils.StatusErrEqual(err, wantErr) {
-		t.Fatalf("TestService/EmptyCall(_, _) = _, %v, want _, %v", err, wantErr)
+		t.Fatalf("TestService/EmptyCall(_, _) = _, %v, want _, %v", status.Convert(err).Proto(), status.Convert(wantErr).Proto())
 	}
 	if _, err := tc.UnaryCall(ctx, &testpb.SimpleRequest{}); err != nil {
 		t.Fatalf("TestService.UnaryCall(%v, _, _, _) = _, %v; want _, <nil>", ctx, err)
@@ -288,7 +284,7 @@ const loadMDKey = "X-Endpoint-Load-Metrics-Bin"
 
 type testLoadParser struct{}
 
-func (*testLoadParser) Parse(md metadata.MD) interface{} {
+func (*testLoadParser) Parse(md metadata.MD) any {
 	vs := md.Get(loadMDKey)
 	if len(vs) == 0 {
 		return nil
@@ -323,7 +319,7 @@ func testDoneLoads(t *testing.T) {
 
 	tc := testgrpc.NewTestServiceClient(ss.CC)
 
-	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
+	ctx, cancel := context.WithTimeout(context.Background(), defaultTestTimeout)
 	defer cancel()
 	if _, err := tc.EmptyCall(ctx, &testpb.Empty{}); err != nil {
 		t.Fatalf("TestService/EmptyCall(_, _) = _, %v, want _, %v", err, nil)
@@ -345,97 +341,6 @@ func testDoneLoads(t *testing.T) {
 	}
 }
 
-const testBalancerKeepAddressesName = "testbalancer-keepingaddresses"
-
-// testBalancerKeepAddresses keeps the addresses in the builder instead of
-// creating SubConns.
-//
-// It's used to test the addresses balancer gets are correct.
-type testBalancerKeepAddresses struct {
-	addrsChan chan []resolver.Address
-}
-
-func newTestBalancerKeepAddresses() *testBalancerKeepAddresses {
-	return &testBalancerKeepAddresses{
-		addrsChan: make(chan []resolver.Address, 10),
-	}
-}
-
-func (testBalancerKeepAddresses) ResolverError(err error) {
-	panic("not implemented")
-}
-
-func (b *testBalancerKeepAddresses) Build(cc balancer.ClientConn, opt balancer.BuildOptions) balancer.Balancer {
-	return b
-}
-
-func (*testBalancerKeepAddresses) Name() string {
-	return testBalancerKeepAddressesName
-}
-
-func (b *testBalancerKeepAddresses) UpdateClientConnState(state balancer.ClientConnState) error {
-	b.addrsChan <- state.ResolverState.Addresses
-	return nil
-}
-
-func (testBalancerKeepAddresses) UpdateSubConnState(sc balancer.SubConn, s balancer.SubConnState) {
-	panic("not used")
-}
-
-func (testBalancerKeepAddresses) Close() {}
-
-func (testBalancerKeepAddresses) ExitIdle() {}
-
-// Make sure that non-grpclb balancers don't get grpclb addresses even if name
-// resolver sends them
-func (s) TestNonGRPCLBBalancerGetsNoGRPCLBAddress(t *testing.T) {
-	r := manual.NewBuilderWithScheme("whatever")
-
-	b := newTestBalancerKeepAddresses()
-	balancer.Register(b)
-
-	cc, err := grpc.Dial(r.Scheme()+":///test.server",
-		grpc.WithTransportCredentials(insecure.NewCredentials()),
-		grpc.WithResolvers(r),
-		grpc.WithDefaultServiceConfig(fmt.Sprintf(`{"loadBalancingConfig": [{"%s":{}}]}`, b.Name())))
-	if err != nil {
-		t.Fatalf("failed to dial: %v", err)
-	}
-	defer cc.Close()
-
-	grpclbAddresses := []resolver.Address{{
-		Addr:       "grpc.lb.com",
-		Type:       resolver.GRPCLB,
-		ServerName: "grpc.lb.com",
-	}}
-
-	nonGRPCLBAddresses := []resolver.Address{{
-		Addr: "localhost",
-		Type: resolver.Backend,
-	}}
-
-	r.UpdateState(resolver.State{
-		Addresses: nonGRPCLBAddresses,
-	})
-	if got := <-b.addrsChan; !reflect.DeepEqual(got, nonGRPCLBAddresses) {
-		t.Fatalf("With only backend addresses, balancer got addresses %v, want %v", got, nonGRPCLBAddresses)
-	}
-
-	r.UpdateState(resolver.State{
-		Addresses: grpclbAddresses,
-	})
-	if got := <-b.addrsChan; len(got) != 0 {
-		t.Fatalf("With only grpclb addresses, balancer got addresses %v, want empty", got)
-	}
-
-	r.UpdateState(resolver.State{
-		Addresses: append(grpclbAddresses, nonGRPCLBAddresses...),
-	})
-	if got := <-b.addrsChan; !reflect.DeepEqual(got, nonGRPCLBAddresses) {
-		t.Fatalf("With both backend and grpclb addresses, balancer got addresses %v, want %v", got, nonGRPCLBAddresses)
-	}
-}
-
 type aiPicker struct {
 	result balancer.PickResult
 	err    error
@@ -488,16 +393,18 @@ func (s) TestAddressAttributesInNewSubConn(t *testing.T) {
 			// Only use the first address.
 			attr := attributes.New(testAttrKey, testAttrVal)
 			addrs[0].Attributes = attr
-			sc, err := bd.ClientConn.NewSubConn([]resolver.Address{addrs[0]}, balancer.NewSubConnOptions{})
+			var sc balancer.SubConn
+			sc, err := bd.ClientConn.NewSubConn([]resolver.Address{addrs[0]}, balancer.NewSubConnOptions{
+				StateListener: func(state balancer.SubConnState) {
+					bd.ClientConn.UpdateState(balancer.State{ConnectivityState: state.ConnectivityState, Picker: &aiPicker{result: balancer.PickResult{SubConn: sc}, err: state.ConnectionError}})
+				},
+			})
 			if err != nil {
 				return err
 			}
 			sc.Connect()
 			return nil
 		},
-		UpdateSubConnState: func(bd *stub.BalancerData, sc balancer.SubConn, state balancer.SubConnState) {
-			bd.ClientConn.UpdateState(balancer.State{ConnectivityState: state.ConnectivityState, Picker: &aiPicker{result: balancer.PickResult{SubConn: sc}, err: state.ConnectionError}})
-		},
 	}
 	stub.Register(attrBalancerName, bf)
 	t.Logf("Registered balancer %s...", attrBalancerName)
@@ -531,7 +438,7 @@ func (s) TestAddressAttributesInNewSubConn(t *testing.T) {
 	t.Log("Created a ClientConn...")
 
 	// The first RPC should fail because there's no address.
-	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
+	ctx, cancel := context.WithTimeout(context.Background(), defaultTestShortTimeout)
 	defer cancel()
 	if _, err := tc.EmptyCall(ctx, &testpb.Empty{}); err == nil || status.Code(err) != codes.DeadlineExceeded {
 		t.Fatalf("EmptyCall() = _, %v, want _, DeadlineExceeded", err)
@@ -543,7 +450,7 @@ func (s) TestAddressAttributesInNewSubConn(t *testing.T) {
 	t.Logf("Pushing resolver state update: %v through the manual resolver", state)
 
 	// The second RPC should succeed.
-	ctx, cancel = context.WithTimeout(context.Background(), time.Second)
+	ctx, cancel = context.WithTimeout(context.Background(), defaultTestTimeout)
 	defer cancel()
 	if _, err := tc.EmptyCall(ctx, &testpb.Empty{}); err != nil {
 		t.Fatalf("EmptyCall() = _, %v, want _, <nil>", err)
@@ -574,18 +481,20 @@ func (s) TestMetadataInAddressAttributes(t *testing.T) {
 				return nil
 			}
 			// Only use the first address.
+			var sc balancer.SubConn
 			sc, err := bd.ClientConn.NewSubConn([]resolver.Address{
 				imetadata.Set(addrs[0], metadata.Pairs(testMDKey, testMDValue)),
-			}, balancer.NewSubConnOptions{})
+			}, balancer.NewSubConnOptions{
+				StateListener: func(state balancer.SubConnState) {
+					bd.ClientConn.UpdateState(balancer.State{ConnectivityState: state.ConnectivityState, Picker: &aiPicker{result: balancer.PickResult{SubConn: sc}, err: state.ConnectionError}})
+				},
+			})
 			if err != nil {
 				return err
 			}
 			sc.Connect()
 			return nil
 		},
-		UpdateSubConnState: func(bd *stub.BalancerData, sc balancer.SubConn, state balancer.SubConnState) {
-			bd.ClientConn.UpdateState(balancer.State{ConnectivityState: state.ConnectivityState, Picker: &aiPicker{result: balancer.PickResult{SubConn: sc}, err: state.ConnectionError}})
-		},
 	}
 	stub.Register(mdBalancerName, bf)
 	t.Logf("Registered balancer %s...", mdBalancerName)
@@ -612,7 +521,7 @@ func (s) TestMetadataInAddressAttributes(t *testing.T) {
 	defer ss.Stop()
 
 	// The RPC should succeed with the expected md.
-	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
+	ctx, cancel := context.WithTimeout(context.Background(), defaultTestTimeout)
 	defer cancel()
 	if _, err := ss.Client.EmptyCall(ctx, &testpb.Empty{}); err != nil {
 		t.Fatalf("EmptyCall() = _, %v, want _, <nil>", err)
@@ -629,7 +538,7 @@ func (s) TestMetadataInAddressAttributes(t *testing.T) {
 // TestServersSwap creates two servers and verifies the client switches between
 // them when the name resolver reports the first and then the second.
 func (s) TestServersSwap(t *testing.T) {
-	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
+	ctx, cancel := context.WithTimeout(context.Background(), defaultTestTimeout)
 	defer cancel()
 
 	// Initialize servers
@@ -685,7 +594,7 @@ func (s) TestServersSwap(t *testing.T) {
 }
 
 func (s) TestWaitForReady(t *testing.T) {
-	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
+	ctx, cancel := context.WithTimeout(context.Background(), defaultTestTimeout)
 	defer cancel()
 
 	// Initialize server
@@ -808,16 +717,18 @@ func (s) TestAuthorityInBuildOptions(t *testing.T) {
 					}
 
 					// Only use the first address.
-					sc, err := bd.ClientConn.NewSubConn([]resolver.Address{addrs[0]}, balancer.NewSubConnOptions{})
+					var sc balancer.SubConn
+					sc, err := bd.ClientConn.NewSubConn([]resolver.Address{addrs[0]}, balancer.NewSubConnOptions{
+						StateListener: func(state balancer.SubConnState) {
+							bd.ClientConn.UpdateState(balancer.State{ConnectivityState: state.ConnectivityState, Picker: &aiPicker{result: balancer.PickResult{SubConn: sc}, err: state.ConnectionError}})
+						},
+					})
 					if err != nil {
 						return err
 					}
 					sc.Connect()
 					return nil
 				},
-				UpdateSubConnState: func(bd *stub.BalancerData, sc balancer.SubConn, state balancer.SubConnState) {
-					bd.ClientConn.UpdateState(balancer.State{ConnectivityState: state.ConnectivityState, Picker: &aiPicker{result: balancer.PickResult{SubConn: sc}, err: state.ConnectionError}})
-				},
 			}
 			balancerName := "stub-balancer-" + test.name
 			stub.Register(balancerName, bf)
@@ -869,38 +780,15 @@ func (s) TestAuthorityInBuildOptions(t *testing.T) {
 	}
 }
 
-// wrappedPickFirstBalancerBuilder builds a custom balancer which wraps an
-// underlying pick_first balancer.
-type wrappedPickFirstBalancerBuilder struct {
-	name string
-}
-
-func (*wrappedPickFirstBalancerBuilder) Build(cc balancer.ClientConn, opts balancer.BuildOptions) balancer.Balancer {
-	builder := balancer.Get(grpc.PickFirstBalancerName)
-	wpfb := &wrappedPickFirstBalancer{
-		ClientConn: cc,
-	}
-	pf := builder.Build(wpfb, opts)
-	wpfb.Balancer = pf
-	return wpfb
-}
-
-func (wbb *wrappedPickFirstBalancerBuilder) Name() string {
-	return wbb.name
-}
-
-// wrappedPickFirstBalancer contains a pick_first balancer and forwards all
-// calls from the ClientConn to it. For state updates from the pick_first
-// balancer, it creates a custom picker which injects arbitrary metadata on a
-// per-call basis.
-type wrappedPickFirstBalancer struct {
-	balancer.Balancer
+// testCCWrapper wraps a balancer.ClientConn and intercepts UpdateState and
+// returns a custom picker which injects arbitrary metadata on a per-call basis.
+type testCCWrapper struct {
 	balancer.ClientConn
 }
 
-func (wb *wrappedPickFirstBalancer) UpdateState(state balancer.State) {
+func (t *testCCWrapper) UpdateState(state balancer.State) {
 	state.Picker = &wrappedPicker{p: state.Picker}
-	wb.ClientConn.UpdateState(state)
+	t.ClientConn.UpdateState(state)
 }
 
 const (
@@ -921,10 +809,10 @@ func (wp *wrappedPicker) Pick(info balancer.PickInfo) (balancer.PickResult, erro
 		return balancer.PickResult{}, err
 	}
 
-	if res.Metatada == nil {
-		res.Metatada = metadata.Pairs(metadataHeaderInjectedByBalancer, metadataValueInjectedByBalancer)
+	if res.Metadata == nil {
+		res.Metadata = metadata.Pairs(metadataHeaderInjectedByBalancer, metadataValueInjectedByBalancer)
 	} else {
-		res.Metatada.Append(metadataHeaderInjectedByBalancer, metadataValueInjectedByBalancer)
+		res.Metadata.Append(metadataHeaderInjectedByBalancer, metadataValueInjectedByBalancer)
 	}
 	return res, nil
 }
@@ -952,10 +840,20 @@ func (s) TestMetadataInPickResult(t *testing.T) {
 	defer ss.Stop()
 	t.Logf("Started test backend at %q", ss.Address)
 
-	name := t.Name() + "wrappedPickFirstBalancer"
-	t.Logf("Registering test balancer with name %q...", name)
-	b := &wrappedPickFirstBalancerBuilder{name: t.Name() + "wrappedPickFirstBalancer"}
-	balancer.Register(b)
+	// Register a test balancer that contains a pick_first balancer and forwards
+	// all calls from the ClientConn to it. For state updates from the
+	// pick_first balancer, it creates a custom picker which injects arbitrary
+	// metadata on a per-call basis.
+	stub.Register(t.Name(), stub.BalancerFuncs{
+		Init: func(bd *stub.BalancerData) {
+			cc := &testCCWrapper{ClientConn: bd.ClientConn}
+			bd.Data = balancer.Get(grpc.PickFirstBalancerName).Build(cc, bd.BuildOptions)
+		},
+		UpdateClientConnState: func(bd *stub.BalancerData, ccs balancer.ClientConnState) error {
+			bal := bd.Data.(balancer.Balancer)
+			return bal.UpdateClientConnState(ccs)
+		},
+	})
 
 	t.Log("Creating ClientConn to test backend...")
 	r := manual.NewBuilderWithScheme("whatever")
@@ -963,7 +861,7 @@ func (s) TestMetadataInPickResult(t *testing.T) {
 	dopts := []grpc.DialOption{
 		grpc.WithTransportCredentials(insecure.NewCredentials()),
 		grpc.WithResolvers(r),
-		grpc.WithDefaultServiceConfig(fmt.Sprintf(`{"loadBalancingConfig": [{"%s":{}}]}`, b.Name())),
+		grpc.WithDefaultServiceConfig(fmt.Sprintf(`{"loadBalancingConfig": [{"%s":{}}]}`, t.Name())),
 	}
 	cc, err := grpc.Dial(r.Scheme()+":///test.server", dopts...)
 	if err != nil {
@@ -1004,3 +902,228 @@ func (s) TestMetadataInPickResult(t *testing.T) {
 		t.Fatalf("Mismatch in custom metadata received at test backend, got: %v, want %v", gotMDVal, wantMDVal)
 	}
 }
+
+// producerTestBalancerBuilder and producerTestBalancer start a producer which
+// makes an RPC before the subconn is READY, then connects the subconn, and
+// pushes the resulting error (expected to be nil) to rpcErrChan.
+type producerTestBalancerBuilder struct {
+	rpcErrChan chan error
+	ctxChan    chan context.Context
+	connect    bool
+}
+
+func (bb *producerTestBalancerBuilder) Build(cc balancer.ClientConn, opts balancer.BuildOptions) balancer.Balancer {
+	return &producerTestBalancer{cc: cc, rpcErrChan: bb.rpcErrChan, ctxChan: bb.ctxChan, connect: bb.connect}
+}
+
+const producerTestBalancerName = "producer_test_balancer"
+
+func (bb *producerTestBalancerBuilder) Name() string { return producerTestBalancerName }
+
+type producerTestBalancer struct {
+	cc         balancer.ClientConn
+	rpcErrChan chan error
+	ctxChan    chan context.Context
+	connect    bool
+}
+
+func (b *producerTestBalancer) UpdateClientConnState(ccs balancer.ClientConnState) error {
+	// Create the subconn, but don't connect it.
+	sc, err := b.cc.NewSubConn(ccs.ResolverState.Addresses, balancer.NewSubConnOptions{})
+	if err != nil {
+		return fmt.Errorf("error creating subconn: %v", err)
+	}
+
+	// Create the producer.  This will call the producer builder's Build
+	// method, which will try to start an RPC in a goroutine.
+	p := &testProducerBuilder{start: grpcsync.NewEvent(), rpcErrChan: b.rpcErrChan, ctxChan: b.ctxChan}
+	sc.GetOrBuildProducer(p)
+
+	// Wait here until the producer is about to perform the RPC, which should
+	// block until connected.
+	<-p.start.Done()
+
+	// Ensure the error chan doesn't get anything on it before we connect the
+	// subconn.
+	select {
+	case err := <-b.rpcErrChan:
+		go func() { b.rpcErrChan <- fmt.Errorf("Got unexpected data on rpcErrChan: %v", err) }()
+	default:
+	}
+
+	if b.connect {
+		// Now we can connect, which will unblock the RPC above.
+		sc.Connect()
+	}
+
+	// The stub server requires a READY picker to be reported, to unblock its
+	// Start method.  We won't make RPCs in our test, so a nil picker is okay.
+	b.cc.UpdateState(balancer.State{ConnectivityState: connectivity.Ready, Picker: nil})
+	return nil
+}
+
+func (b *producerTestBalancer) ResolverError(err error) {
+	panic(fmt.Sprintf("Unexpected resolver error: %v", err))
+}
+
+func (b *producerTestBalancer) UpdateSubConnState(balancer.SubConn, balancer.SubConnState) {}
+func (b *producerTestBalancer) Close()                                                     {}
+
+type testProducerBuilder struct {
+	start      *grpcsync.Event
+	rpcErrChan chan error
+	ctxChan    chan context.Context
+}
+
+func (b *testProducerBuilder) Build(cci any) (balancer.Producer, func()) {
+	c := testgrpc.NewTestServiceClient(cci.(grpc.ClientConnInterface))
+	// Perform the RPC in a goroutine instead of during build because the
+	// subchannel's mutex is held here.
+	go func() {
+		ctx := <-b.ctxChan
+		b.start.Fire()
+		_, err := c.EmptyCall(ctx, &testpb.Empty{})
+		b.rpcErrChan <- err
+	}()
+	return nil, func() {}
+}
+
+// TestBalancerProducerBlockUntilReady tests that we get no RPC errors from
+// producers when subchannels aren't ready.
+func (s) TestBalancerProducerBlockUntilReady(t *testing.T) {
+	// rpcErrChan is given to the LB policy to report the status of the
+	// producer's one RPC.
+	ctxChan := make(chan context.Context, 1)
+	ctx, cancel := context.WithTimeout(context.Background(), defaultTestTimeout)
+	defer cancel()
+	ctxChan <- ctx
+
+	rpcErrChan := make(chan error)
+	balancer.Register(&producerTestBalancerBuilder{rpcErrChan: rpcErrChan, ctxChan: ctxChan, connect: true})
+
+	ss := &stubserver.StubServer{
+		EmptyCallF: func(ctx context.Context, in *testpb.Empty) (*testpb.Empty, error) {
+			return &testpb.Empty{}, nil
+		},
+	}
+
+	// Start the server & client with the test producer LB policy.
+	svcCfg := fmt.Sprintf(`{"loadBalancingConfig": [{"%s":{}}]}`, producerTestBalancerName)
+	if err := ss.Start(nil, grpc.WithDefaultServiceConfig(svcCfg)); err != nil {
+		t.Fatalf("Error starting testing server: %v", err)
+	}
+	defer ss.Stop()
+
+	// Receive the error from the producer's RPC, which should be nil.
+	if err := <-rpcErrChan; err != nil {
+		t.Fatalf("Received unexpected error from producer RPC: %v", err)
+	}
+}
+
+// TestBalancerProducerHonorsContext tests that producers that perform RPC get
+// context errors correctly.
+func (s) TestBalancerProducerHonorsContext(t *testing.T) {
+	// rpcErrChan is given to the LB policy to report the status of the
+	// producer's one RPC.
+	ctxChan := make(chan context.Context, 1)
+	ctx, cancel := context.WithTimeout(context.Background(), defaultTestTimeout)
+	ctxChan <- ctx
+
+	rpcErrChan := make(chan error)
+	balancer.Register(&producerTestBalancerBuilder{rpcErrChan: rpcErrChan, ctxChan: ctxChan, connect: false})
+
+	ss := &stubserver.StubServer{
+		EmptyCallF: func(ctx context.Context, in *testpb.Empty) (*testpb.Empty, error) {
+			return &testpb.Empty{}, nil
+		},
+	}
+
+	// Start the server & client with the test producer LB policy.
+	svcCfg := fmt.Sprintf(`{"loadBalancingConfig": [{"%s":{}}]}`, producerTestBalancerName)
+	if err := ss.Start(nil, grpc.WithDefaultServiceConfig(svcCfg)); err != nil {
+		t.Fatalf("Error starting testing server: %v", err)
+	}
+	defer ss.Stop()
+
+	cancel()
+
+	// Receive the error from the producer's RPC, which should be canceled.
+	if err := <-rpcErrChan; status.Code(err) != codes.Canceled {
+		t.Fatalf("RPC error: %v; want status.Code(err)=%v", err, codes.Canceled)
+	}
+}
+
+// TestSubConnShutdown confirms that the Shutdown method on subconns and
+// RemoveSubConn method on ClientConn properly initiates subconn shutdown.
+func (s) TestSubConnShutdown(t *testing.T) {
+	ctx, cancel := context.WithTimeout(context.Background(), defaultTestTimeout)
+	defer cancel()
+
+	testCases := []struct {
+		name     string
+		shutdown func(cc balancer.ClientConn, sc balancer.SubConn)
+	}{{
+		name: "ClientConn.RemoveSubConn",
+		shutdown: func(cc balancer.ClientConn, sc balancer.SubConn) {
+			cc.RemoveSubConn(sc)
+		},
+	}, {
+		name: "SubConn.Shutdown",
+		shutdown: func(_ balancer.ClientConn, sc balancer.SubConn) {
+			sc.Shutdown()
+		},
+	}}
+
+	for _, tc := range testCases {
+		t.Run(tc.name, func(t *testing.T) {
+			gotShutdown := grpcsync.NewEvent()
+
+			bf := stub.BalancerFuncs{
+				UpdateClientConnState: func(bd *stub.BalancerData, ccs balancer.ClientConnState) error {
+					var sc balancer.SubConn
+					opts := balancer.NewSubConnOptions{
+						StateListener: func(scs balancer.SubConnState) {
+							switch scs.ConnectivityState {
+							case connectivity.Connecting:
+								// Ignored.
+							case connectivity.Ready:
+								tc.shutdown(bd.ClientConn, sc)
+							case connectivity.Shutdown:
+								gotShutdown.Fire()
+							default:
+								t.Errorf("got unexpected state %q in listener", scs.ConnectivityState)
+							}
+						},
+					}
+					sc, err := bd.ClientConn.NewSubConn(ccs.ResolverState.Addresses, opts)
+					if err != nil {
+						return err
+					}
+					sc.Connect()
+					// Report the state as READY to unblock ss.Start(), which waits for ready.
+					bd.ClientConn.UpdateState(balancer.State{ConnectivityState: connectivity.Ready})
+					return nil
+				},
+			}
+
+			testBalName := "shutdown-test-balancer-" + tc.name
+			stub.Register(testBalName, bf)
+			t.Logf("Registered balancer %s...", testBalName)
+
+			ss := &stubserver.StubServer{}
+			if err := ss.Start(nil, grpc.WithDefaultServiceConfig(
+				fmt.Sprintf(`{ "loadBalancingConfig": [{"%v": {}}] }`, testBalName),
+			)); err != nil {
+				t.Fatalf("Error starting endpoint server: %v", err)
+			}
+			defer ss.Stop()
+
+			select {
+			case <-gotShutdown.Done():
+				// Success
+			case <-ctx.Done():
+				t.Fatalf("Timed out waiting for gotShutdown to be fired.")
+			}
+		})
+	}
+}
